/*********************

Code package includes modified libraries for scanning PID's.
Cycle power to Arduino board to restart scan
Observice results at Tools>Serial Monitor
Test of 127 PIDs at 19200 baud takes about 2 minutes

Default currently set to measure PID range 2F to 3A in order to easily test the Smart Fortwo ED3 EBS2 HV battery current sensor
A good sensor will show data on PIDs 30,34,35,36,37,38,39 A failed sensor or no power will display Scan Complete without displaying any data
If all tested PIDs show data of 00 then the connections to the LIN driver are not providing the necessary echo.

LIN_HW error: 0x4 is common in a scan when the reply is less than 8 bytes.
Packets with both timeout and checksum error are not displayed, since those indicate no response. 
Error codes:
NO_ERROR      = 0x00,                     //!< no error
ERROR_STATE   = 0x01,                     //!< error in LIN state machine
ERROR_ECHO    = 0x02,                     //!< error reading LIN echo
ERROR_TIMEOUT = 0x04,                     //!< frame timeout error
ERROR_CHK     = 0x08,                     //!< LIN checksum error
ERROR_MISC    = 0x80                      //!< misc error, should not occur

Modified Example code for LIN master node with blocking operation using HardwareSerial
This code runs a LIN master node in blocking operation using HardwareSerial interface

Original sources from https://github.com/gicking/LIN_master_portable_Arduino
This version by Otmar at CafeElectric.com. 10/8/2023

MIT License

Copyright (c) 2023 Otmar Ebenhoech

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

------------------------
Hardware setup:
Tested on a Arduino Mega 2560 attached to a FST T151 LIN driver board with TJA1021, TJA1020 should work too. 

T151 to Mega 2560 board connections:
GND > GND
TX > RX3 (pin 15)
RX > TX3 (pin 14)

T151 to Device Under Test (DUT) LIN connections
GND > Chassis ground
12V > 12V on DUT plus power supply
LIN > DUT Lin connection

T151 SLP and INH are not connected
------------------------

Supported (=successfully tested) boards:
 - Arduino Mega 2560      https://store.arduino.cc/products/arduino-mega-2560-rev3
 - T151                   Amazon affiliate link https://amzn.to/45lYuwS
**********************/

// include files
#include "LIN_master_HardwareSerial.h"

#define SCAN_ONCE 1  // comment out this line for endless scanning

#define START_PID 0x2F  // Starting PID to send 00 covers all possibilities
#define END_PID 0x3A  // End of scan PID, 7F covers all possibilities

// pin to demonstrate background operation
#define PIN_TOGGLE 30

// indicate LIN return status
#define PIN_ERROR 32

// pause between LIN frames
#define LIN_PAUSE 5

uint8_t PID;

// setup LIN node
LIN_Master_HardwareSerial LIN(Serial3, "LIN_HW");  // parameter: HW-interface, name Serial3 indicates using pins TX3 and RX3 on the Arduino Mega 2660


// call once
void setup() {


  PID = START_PID;  // set first PID to scan

  // indicate background operation
  pinMode(PIN_TOGGLE, OUTPUT);

  // indicate LIN status via pin
  pinMode(PIN_ERROR, OUTPUT);

  // open LIN interface
  // int baud = 9600;   // set LIN baud rate here
  int baud = 19200;  // set LIN baud rate here
  LIN.begin(baud);

  // for user interaction via console at 115200 baud
  Serial.begin(115200);
  while (!Serial)  // wait for interface to initialize
    ;
  Serial.println("");
  Serial.print("Starting scan at PID 0x");
  if (PID < 0x10)       // if value is only one digit
    Serial.print('0');  // add a leading zero
  Serial.print(PID, HEX);
  Serial.print(" at ");
  Serial.print(baud);  // display LIN baud rate
  Serial.println(" baud");

}  // setup()


// call repeatedly
void loop() {
  static uint32_t tStart;
  LIN_Master::frame_t Type;
  uint8_t Id;
  uint8_t NumData;
  uint8_t Data[8];
  LIN_Master::error_t error;


  // reset state machine & error
  LIN.resetStateMachine();
  LIN.resetError();

  // wait a bit. Toggle pin to show CPU load
  tStart = millis();
  while (millis() - tStart < LIN_PAUSE) {
    digitalWrite(PIN_TOGGLE, !digitalRead(PIN_TOGGLE));
  }

  // send/receive slave response frame and get result immediately
  error = LIN.receiveSlaveResponseBlocking(LIN_Master::LIN_V2, PID, 8, Data);  //

  // indicate status via pin
  digitalWrite(PIN_ERROR, error);

  // get frame data
  LIN.getFrame(Type, Id, NumData, Data);


// print result
#if !defined(SKIP_CONSOLE)
  if ((error != 0xC))  // if it's not timeout and checksum error (8+4)
  {                    // Print the output
    Serial.print("PID: ");
    if (PID < 0x10)          // if value is only one digit
      Serial.print('0');     // add a leading zero
    Serial.print(PID, HEX);  // Print the current PID
    Serial.print("\tData 0-7\t:");

    for (uint8_t i = 0; (i < NumData); i++)  // Print data bytes recieved
    {
      if ((int)Data[i] < 0x10)          // if value is only one digit
        Serial.print('0');              // add a leading zero
      Serial.print((int)Data[i], HEX);  // print data value in Hex
      if (i < (NumData - 1))            // if it's not the end of line
        Serial.print(",");              // print a comma to seperate data
    }
    Serial.print("\tTime: ");  // at the end add time and error information
    Serial.print(millis());
    Serial.print("\t");
    Serial.print(LIN.nameLIN);
    Serial.print(" error: 0x");
    if (error < 0x10)            // if value is only one digit
      Serial.print('0');         // add a leading zero
    Serial.println(error, HEX);  // new line after data is printed
  };
#endif  // SKIP_CONSOLE

  // reset state machine & error
  LIN.resetStateMachine();
  LIN.resetError();

  // wait a bit. Toggle pin to show CPU load
  tStart = millis();
  while (millis() - tStart < LIN_PAUSE) {
    digitalWrite(PIN_TOGGLE, !digitalRead(PIN_TOGGLE));
  }


  PID++; // increment PID Value

  if (PID >= END_PID) { // We've come to the end of the scan
#if defined SCAN_ONCE // Do we continue? 
    Serial.print("Scan Complete at PID 0x");
    if (PID < 0x10)      // if value is only one digit
    Serial.print('0');  // add a leading zero
  Serial.print(PID, HEX);
    while (1) { /* spin and do nothing */
    }           // End of scan.
#else  // if we continue, then start over
    PID = START_PID;
#endif
  }
}  // loop()
